/*
 * parse.cpp: A plugin for the Video Disk Recorder
 *
 * See the README file for copyright information and how to reach the author.
 *
 */

#include <string.h>
#include <stdlib.h>
#include <limits.h>

#include <fcntl.h>
#include <unistd.h>

#include "parse.h"

void cXMLTVEvent::SetTitle(const char *Title)
{
    title = strcpyrealloc(title, Title);
    title = compactspace(title);
}

void cXMLTVEvent::SetShortText(const char *ShortText)
{
    shorttext=strcpyrealloc(shorttext,ShortText);
    shorttext=compactspace(shorttext);
}

void cXMLTVEvent::SetDescription(const char *Description)
{
    description = strcpyrealloc(description, Description);
    description = compactspace(description);
}

void cXMLTVEvent::Clear()
{
    if (title)
    {
        free(title);
        title=NULL;
    }
    if (shorttext)
    {
        free(shorttext);
        shorttext=NULL;
    }
    if (description)
    {
        free(description);
        description=NULL;
    }
    starttime = 0;
    duration = 0;
    eventid=0;
}

cXMLTVEvent::cXMLTVEvent()
{
    title=NULL;
    shorttext=NULL;
    description=NULL;
    starttime = 0;
    duration = 0;
}

cXMLTVEvent::~cXMLTVEvent()
{
    Clear();
}

// -------------------------------------------------------

time_t cParse::ConvertXMLTVTime2UnixTime(char *xmltvtime)
{
    time_t offset=0;
    if (!xmltvtime) return (time_t) 0;
    char *withtz=strchr(xmltvtime,' ');
    int len;
    if (withtz)
    {
        len=strlen(xmltvtime)-(withtz-xmltvtime)-1;
        *withtz=':';
        if ((withtz[1]=='+') || (withtz[1]=='-'))
        {
            if (len==5)
            {
                int val=atoi(&withtz[1]);
                int h=val/100;
                int m=val-(h*100);
                offset=h*3600+m*60;
                setenv("TZ",":UTC",1);
            }
            else
            {
                setenv("TZ",":UTC",1);
            }
        }
        else
        {
            if (len>2)
            {
                setenv("TZ",withtz,1);
            }
            else
            {
                setenv("TZ",":UTC",1);
            }
        }
    }
    else
    {
        withtz=&xmltvtime[strlen(xmltvtime)];
        setenv("TZ",":UTC",1);
    }
    tzset();

    len=withtz-xmltvtime;
    if (len<4)
    {
        unsetenv("TZ");
        tzset();
        return (time_t) 0;
    }
    len-=2;
    char fmt[]="%Y%m%d%H%M%S";
    fmt[len]=0;

    struct tm tm;
    memset(&tm,0,sizeof(tm));
    if (!strptime(xmltvtime,fmt,&tm))
    {
        unsetenv("TZ");
        tzset();
        return (time_t) 0;
    }
    if (tm.tm_mday==0) tm.tm_mday=1;
    time_t ret=mktime(&tm);
    ret-=offset;
    unsetenv("TZ");
    tzset();
    return ret;
}

cEvent *cParse::SearchEvent(cSchedule* schedule, cXMLTVEvent *xevent)
{
    if (!xevent) return NULL;
    if (!xevent->Duration()) return NULL;
    if (!xevent->StartTime()) return NULL;
    if (!xevent->Title()) return NULL;
    cEvent *f=NULL;

    time_t start=xevent->StartTime();
    if (!xevent->EventID())
    {
        xevent->SetEventID(DoSum((u_long) start,xevent->Title(),strlen(xevent->Title())));
    }

    // try to find an event,
    // 1st with StartTime
    f=(cEvent *) schedule->GetEvent((tEventID) 0,start);
    if (f) return f;
    // 2nd with our own EventID
    if (xevent->EventID()) f=(cEvent *) schedule->GetEvent((tEventID) xevent->EventID());
    if (f) return f;
    // 3rd with StartTime +/- WaitTime
    int maxdiff=INT_MAX;
    int eventTimeDiff=xevent->Duration()/4;
    if (eventTimeDiff<600) eventTimeDiff=600;

    for (cEvent *p = schedule->Events()->First(); p; p = schedule->Events()->Next(p))
    {
        if (!strcmp(p->Title(),xevent->Title()))
        {
            // found event with same title
            int diff=abs((int) difftime(p->StartTime(),start));
            if (diff<=eventTimeDiff)
            {
                if (diff<=maxdiff)
                {
                    f=p;
                    maxdiff=diff;
                }
            }
        }
    }
    return f;
}

bool cParse::PutEvent(cSchedule* schedule, cEvent *event, cXMLTVEvent *xevent, cEPGMapping *map)
{
    if (!schedule) return false;
    if (!xevent) return false;
    if (!map) return false;

    time_t start;
    if (!event)
    {
        if ((map->Flags() & OPT_APPEND)==OPT_APPEND)
        {
            start=xevent->StartTime();
            event=new cEvent(xevent->EventID());
            if (!event) return false;
            event->SetStartTime(start);
            event->SetDuration(xevent->Duration());
            event->SetTitle(xevent->Title());
            event->SetVersion(0);
            schedule->AddEvent(event);
            dsyslog("xmltv2vdr: '%s' adding event '%s' @%s",name,xevent->Title(),ctime(&start));
        }
        else
        {
            return true;
        }
    }
    else
    {
        start=event->StartTime();
        dsyslog("xmltv2vdr: '%s' changing event '%s' @%s",name,event->Title(),ctime(&start));
    }
    if ((map->Flags() & USE_SHORTTEXT)==USE_SHORTTEXT)
        event->SetShortText(xevent->ShortText());
    if ((map->Flags() & USE_LONGTEXT)==USE_LONGTEXT)
        event->SetDescription(xevent->Description());
    event->SetTableID(0); // prevent EIT EPG to update this event
    return true;
}

u_long cParse::DoSum(u_long sum, const char *buf, int nBytes)
{
    int nleft=nBytes;
    u_short *w = (u_short*)buf;

    while (nleft > 1)
    {
        sum += *w++;
        nleft -= 2;
    }

    if (nleft == 1)
    {
        u_short answer = 0;
        *(u_char*)(&answer) = *(u_char*)w;
        sum += answer;
    }
    return sum;
}

bool cParse::FetchEvent(xmlNodePtr enode)
{
    xmlNodePtr node=enode->xmlChildrenNode;
    while (node)
    {
        if (node->type==XML_ELEMENT_NODE)
        {
            if ((!xmlStrcasecmp(node->name, (const xmlChar *) "title")))
            {
                // what to do with attribute lang?
                xmlChar *content=xmlNodeListGetString(node->doc,node->xmlChildrenNode,1);
                if (content)
                {
                    xevent.SetTitle(conv->Convert((const char *) content));
                    xmlFree(content);
                }
            }
            else if ((!xmlStrcasecmp(node->name, (const xmlChar *) "sub-title")))
            {
                // what to do with attribute lang?
                xmlChar *content=xmlNodeListGetString(node->doc,node->xmlChildrenNode,1);
                if (content)
                {
                    xevent.SetShortText(conv->Convert((const char *) content));
                    xmlFree(content);
                }
            }
            else if ((!xmlStrcasecmp(node->name, (const xmlChar *) "desc")))
            {
                // what to do with attribute lang?
                xmlChar *content=xmlNodeListGetString(node->doc,node->xmlChildrenNode,1);
                if (content)
                {
                    xevent.SetDescription(conv->Convert((const char *) content));
                    xmlFree(content);
                }
            }
            else if ((!xmlStrcasecmp(node->name, (const xmlChar *) "country")))
            {
#if 0
                xmlChar *content=xmlNodeListGetString(node->doc,node->xmlChildrenNode,1);
                if (content)
                {
                    xmlFree(content);
                }
#endif
            }
            else if ((!xmlStrcasecmp(node->name, (const xmlChar *) "date")))
            {
#if 0
                xmlChar *content=xmlNodeListGetString(node->doc,node->xmlChildrenNode,1);
                if (content)
                {
                    xmlFree(content);
                }
#endif
            }
            else if ((!xmlStrcasecmp(node->name, (const xmlChar *) "category")))
            {
                // attribute lang
            }
            else if ((!xmlStrcasecmp(node->name, (const xmlChar *) "credits")))
            {
            }
            else if ((!xmlStrcasecmp(node->name, (const xmlChar *) "rating")))
            {
            }
            else if ((!xmlStrcasecmp(node->name, (const xmlChar *) "review")))
            {
            }
            else if ((!xmlStrcasecmp(node->name, (const xmlChar *) "video")))
            {
            }
            else if ((!xmlStrcasecmp(node->name, (const xmlChar *) "audio")))
            {
            }
            else
            {
                dsyslog("xmltv2vdr: '%s' unknown element %s",name,node->name);
            }
        }
        node=node->next;
    }
    return true;
}

cEPGMapping *cParse::EPGMapping(const char *ChannelName)
{
    if (!maps->Count()) return NULL;
    for (cEPGMapping *map=maps->First(); map; map=maps->Next(map))
    {
        if (!strcmp(map->ChannelName(),ChannelName)) return map;
    }
    return NULL;
}

bool cParse::Process(char *buffer, int bufsize, int days)
{
    if (!buffer) return false;
    if (!bufsize) return false;
    if (!days) return false;

    xmlDocPtr xmltv;
    xmltv=xmlReadMemory(buffer,bufsize,NULL,NULL,0);
    if (!xmltv) return false;

    xmlNodePtr rootnode=xmlDocGetRootElement(xmltv);
    if (!rootnode)
    {
        xmlFreeDoc(xmltv);
        return false;
    }

    time_t begin=time(NULL);
    time_t end=begin+86000*days;
    xmlNodePtr node=rootnode->xmlChildrenNode;
    while (node)
    {
        if (node->type==XML_ELEMENT_NODE)
        {
            if ((!xmlStrcasecmp(node->name, (const xmlChar *) "programme")))
            {
                xmlChar *start,*stop;
                time_t starttime=(time_t) 0;
                time_t stoptime=(time_t) 0;
                start=xmlGetProp(node,(const xmlChar *) "start");
                if (start)
                {
                    starttime=ConvertXMLTVTime2UnixTime((char *) start);
                    if (starttime)
                    {
                        stop=xmlGetProp(node,(const xmlChar *) "stop");
                        if (stop)
                        {
                            stoptime=ConvertXMLTVTime2UnixTime((char *) stop);
                        }
                    }
                }

                if (starttime && (starttime>begin) && (starttime<end))
                {
                    xevent.Clear();

                    xmlChar *eventid;
                    eventid=xmlGetProp(node,(const xmlChar *) "eventid");
                    if (eventid)
                    {
                        xevent.SetEventID(strtoul((const char *) eventid,NULL,2));
                    }

                    xmlChar *channelid;
                    channelid=xmlGetProp(node,(const xmlChar *) "channelid");
                    cEPGMapping *map=NULL;
                    if (channelid && (map=EPGMapping((const char *) channelid)))
                    {
                        xevent.SetStartTime(starttime);
                        if (stoptime) xevent.SetDuration(stoptime-starttime);
                        FetchEvent(node);

                        cSchedulesLock *schedulesLock = new cSchedulesLock(true,2000); // to be safe ;)
                        const cSchedules *schedules = cSchedules::Schedules(*schedulesLock);
                        if (schedules)
                        {
                            for (int i=0; i<map->NumChannelIDs(); i++)
                            {
                                bool addevents=false;
                                if ((map->Flags() & OPT_APPEND)==OPT_APPEND) addevents=true;

                                cChannel *channel=Channels.GetByChannelID(map->ChannelIDs()[i]);
                                cSchedule* schedule = (cSchedule *) schedules->GetSchedule(channel,addevents);
                                if (schedule)
                                {
                                    cEvent *event=NULL;
                                    if ((event=SearchEvent(schedule,&xevent)))
                                    {
                                        PutEvent(schedule,event,&xevent,map);
                                    }
                                    else
                                    {
                                        if (addevents) PutEvent(schedule,event,&xevent,map);
                                    }
                                }
                            }
                        }
                        delete schedulesLock;
                    }
                }
            }
        }
        node=node->next;
    }
    xmlFreeDoc(xmltv);
    return true;
}

cParse::cParse(const char *Name, cEPGMappings *Maps)
{
#if VDRVERSNUM < 10701 || defined(__FreeBSD__)
    conv = new cCharSetConv("UTF-8",cCharSetConv::SystemCharacterTable() ?
                            cCharSetConv::SystemCharacterTable() : "UTF-8");
#else
    conv = new cCharSetConv("UTF-8",NULL);
#endif
    name=strdup(Name);
    maps=Maps;
}

cParse::~cParse()
{
    delete conv;
    free(name);
}
