/*
 * xmltv2vdr.cpp: A plugin for the Video Disk Recorder
 *
 * See the README file for copyright information and how to reach the author.
 *
 */

#include <vdr/plugin.h>
#include <string.h>
#include <sys/wait.h>
#include "xmltv2vdr.h"
#include "extpipe.h"
#include "setup.h"

#if __GNUC__ > 3
#define UNUSED(v) UNUSED_ ## v __attribute__((unused))
#else
#define UNUSED(x) x
#endif

// -------------------------------------------------------------

cEPGExecute::cEPGExecute(const char *Name,cEPGMappings *Maps):cThread(Name)
{
    dsyslog("xmltv2vdr: added epgsource '%s'",Name);
    name=strdup(Name);
    pipe=false;
    daysinadvance=0;
    ready2parse=ReadConfig();
    parse=new cParse(Name, Maps);
}

cEPGExecute::~cEPGExecute()
{
    Stop();
    free((void*) name);
    if (parse) delete parse;
}

bool cEPGExecute::ReadConfig()
{
    dsyslog("xmltv2vdr: reading config of epgsource '%s'",name);
    char *fname=NULL;
    if (asprintf(&fname,"%s/%s",EPGSOURCES,name)!=-1)
    {
        FILE *f=fopen(fname,"r+");
        if (f)
        {
            size_t lsize;
            char *line=NULL;
            int linenr=1;
            while (getline(&line,&lsize,f)!=-1)
            {
                if (linenr==1)
                {
                    if (!strncmp(line,"pipe",4))
                    {
                        dsyslog("xmltv2vdr: '%s' is providing data through a pipe",name);
                        pipe=true;
                    }
                    else
                    {
                        dsyslog("xmltv2vdr: '%s' is providing data through a file",name);
                        pipe=false;
                    }
                }
                if (linenr==2)
                {
                    char *semicolon=strchr(line,';');
                    if (semicolon)
                    {
                        *semicolon=0;
                        semicolon++;
                        daysinadvance=atoi(line);
                        daysmax=atoi(semicolon);
                        dsyslog("xmltv2vdr: '%s' days=%i/%i",name,daysinadvance,daysmax);
                    }
                }
                if (linenr>2)
                {
                    // channels
                    char *semicolon=strchr(line,';');
                    if (semicolon)
                    {
                        *semicolon=0;
                        if (line[0]=='*')
                        {
                            char *buf=NULL;
                            if (asprintf(&buf,"%s*",&line[1])!=-1)
                            {
                                dsyslog("xmltv2vdr: '%s' channel=%s*",name,&line[1]);
                                channels.Append(buf);
                            }
                        }
                        else
                        {
                            channels.Append(strdup(line));
                        }
                    }
                }
                linenr++;
            }
            if (line) free(line);
            channels.Sort();
            fclose(f);
            free(fname);
            return true;
        }
        else
        {
            esyslog("xmltv2vdr: '%s' cannot read config file",name);
        }
        free(fname);
    }
    return false;
}

void cEPGExecute::Action()
{
    if (!ready2parse) return;
    if (!parse) return;
    char *result=NULL;
    int l=0;

    if (pipe)
    {
        cExtPipe p;
        if (p.Open(name,"r"))
        {
            dsyslog("xmltv2vdr: executing epgsource '%s'",name);
            int c;
            while ((c=fgetc(p))!=EOF)
            {
                if (l%20==0) result=(char *) realloc(result, l+21);
                result[l++]=c;
            }
            int status;
            if (p.Close(status)>0)
            {
                int returncode=WEXITSTATUS(status);
                if ((!returncode) && (result))
                {
                    dsyslog("xmltv2vdr: parsing output of '%s'",name);
                    result[l]=0;
                    if (!parse->Process(result,l,daysinadvance))
                    {
                        esyslog("xmltv2vdr: failed to parse output of '%s'",name);
                    }
                }
                else
                {
                    esyslog("xmltv2vdr: epgsource '%s' returned with %i",name,returncode);
                }
            }
            if (result) free(result);
        }
        else
        {
            esyslog("xmltv2vdr: failed to open pipe for '%s'",name);
        }
    }
    else
    {
        char *fname=NULL;
        if (asprintf(&fname,"%s/%s.xmltv",EPGSOURCES,name)!=-1)
        {
            int fd=open(fname,O_RDONLY);
            if (fd!=-1)
            {
                struct stat statbuf;
                if (fstat(fd,&statbuf)!=-1)
                {
                    l=statbuf.st_size;
                    result=(char *) malloc(l+1);
                    if (result)
                    {
                        if (read(fd,result,statbuf.st_size)==statbuf.st_size)
                        {
                            parse->Process(result,l,daysinadvance);
                        }
                        free(result);
                    }
                }
                close(fd);
            }
            free(fname);
        }
    }
}

// -------------------------------------------------------------

cEPGSource::cEPGSource(const char *Name, cEPGMappings *Maps):exec(Name,Maps)
{
    name=strdup(Name);
}

cEPGSource::~cEPGSource()
{
    dsyslog("xmltv2vdr: removed epgsource '%s'",name);
    free((void *) name);
}

bool cEPGSource::Execute()
{
    if (exec.Active()) return false;
    exec.Start();
    return true;
}

// -------------------------------------------------------------

bool cPluginXmltv2vdr::epgsourceexists(const char *name)
{
    if (!epgsources.Count()) return false;
    for (cEPGSource *epgs=epgsources.First(); epgs; epgs=epgsources.Next(epgs))
    {
        if (!strcmp(epgs->Name(),name)) return true;
    }
    return false;
}

void cPluginXmltv2vdr::removeepgmappings()
{
    cEPGMapping *maps;
    while ((maps=epgmappings.Last())!=NULL)
    {
        epgmappings.Del(maps);
    }
}

void cPluginXmltv2vdr::removeepgsources()
{
    cEPGSource *epgs;
    while ((epgs=epgsources.Last())!=NULL)
    {
        epgsources.Del(epgs);
    }
}

bool cPluginXmltv2vdr::executeepgsources()
{
    bool ret=false;
    for (cEPGSource *epgs=epgsources.First(); epgs; epgs=epgsources.Next(epgs))
    {
        ret|=epgs->Execute();
    }
    return ret;
}

bool cPluginXmltv2vdr::EPGMappingExists(tChannelID ChannelID, const char *Channel2Ignore)
{
    if (!epgmappings.Count()) return false;
    for (cEPGMapping *maps=epgmappings.First(); maps; maps=epgmappings.Next(maps))
    {
        if (Channel2Ignore && !strcmp(maps->ChannelName(),Channel2Ignore)) continue;
        for (int i=0; i<maps->NumChannelIDs(); i++)
        {
            if (maps->ChannelIDs()[i]==ChannelID) return true;
        }
    }
    return false;
}

void cPluginXmltv2vdr::EPGMappingReplace(cEPGMapping *newMapping)
{
    if (!newMapping) return;
    cEPGMapping *map=EPGMapping(newMapping->ChannelName());
    if (!map)
    {
        map=new cEPGMapping(*newMapping);
        epgmappings.Add(map);
    }
    else
    {
        map->ChangeFlags(newMapping->Flags());
        map->ReplaceChannels(newMapping->NumChannelIDs(),newMapping->ChannelIDs());
    }
}

cEPGMapping *cPluginXmltv2vdr::EPGMapping(const char *ChannelName)
{
    if (!epgmappings.Count()) return NULL;
    for (cEPGMapping *maps=epgmappings.First(); maps; maps=epgmappings.Next(maps))
    {
        if (!strcmp(maps->ChannelName(),ChannelName)) return maps;
    }
    return NULL;
}

void cPluginXmltv2vdr::ReadInEPGSources(bool Reload)
{
    if (Reload) removeepgsources();
    DIR *dir=opendir(EPGSOURCES);
    if (!dir) return;
    struct dirent *dirent;
    while (dirent=readdir(dir))
    {
        if (strchr(&dirent->d_name[0],'.')) continue;
        if (!epgsourceexists(dirent->d_name))
        {
            char *path=NULL;
            if (asprintf(&path,"%s/%s",EPGSOURCES,dirent->d_name)!=-1)
            {
                if (access(path,R_OK|W_OK)!=-1)
                {
                    epgsources.Add(new cEPGSource(dirent->d_name,&epgmappings));
                }
                else
                {
                    esyslog("xmltv2vdr: cannot access config file for '%s'",dirent->d_name);
                }
                free(path);
            }
        }
    }
    closedir(dir);
}

cPluginXmltv2vdr::cPluginXmltv2vdr(void)
{
    // Initialize any member variables here.
    // DON'T DO ANYTHING ELSE THAT MAY HAVE SIDE EFFECTS, REQUIRE GLOBAL
    // VDR OBJECTS TO EXIST OR PRODUCE ANY OUTPUT!
}

cPluginXmltv2vdr::~cPluginXmltv2vdr()
{
    // Clean up after yourself!
}

const char *cPluginXmltv2vdr::CommandLineHelp(void)
{
    // Return a string that describes all known command line options.
    return NULL;
}

bool cPluginXmltv2vdr::ProcessArgs(int UNUSED(argc), char *UNUSED(argv[]))
{
    // Implement command line argument processing here if applicable.
    return true;
}

bool cPluginXmltv2vdr::Initialize(void)
{
    // Initialize any background activities the plugin shall perform.
    return true;
}

bool cPluginXmltv2vdr::Start(void)
{
    // Start any background activities the plugin shall perform.
    ReadInEPGSources();
    return true;
}

void cPluginXmltv2vdr::Stop(void)
{
    // Stop any background activities the plugin is performing.
    removeepgsources();
    removeepgmappings();
}

void cPluginXmltv2vdr::Housekeeping(void)
{
    // Perform any cleanup or other regular tasks.
}

void cPluginXmltv2vdr::MainThreadHook(void)
{
    // Perform actions in the context of the main program thread.
    // WARNING: Use with great care - see PLUGINS.html!
}

cString cPluginXmltv2vdr::Active(void)
{
    // Return a message string if shutdown should be postponed
    return NULL;
}

time_t cPluginXmltv2vdr::WakeupTime(void)
{
    // Return custom wakeup time for shutdown script
    return 0;
}

const char *cPluginXmltv2vdr::MainMenuEntry(void)
{
    // Return a main menu entry
    return NULL;
}

cOsdObject *cPluginXmltv2vdr::MainMenuAction(void)
{
    // Perform the action when selected from the main VDR menu.
    return NULL;
}

cMenuSetupPage *cPluginXmltv2vdr::SetupMenu(void)
{
    // Return a setup menu in case the plugin supports one.
    return new cMenuSetupXmltv2vdr(this);
}

bool cPluginXmltv2vdr::SetupParse(const char *Name, const char *Value)
{
    // Parse your own setup parameters and store their values.
    if (!strncasecmp(Name,"channel",7))
    {
        if (strlen(Name)<10) return false;

        cEPGMapping *map = new cEPGMapping(&Name[8],Value);
        epgmappings.Add(map);
    }
    else return false;
    return true;
}

bool cPluginXmltv2vdr::Service(const char *UNUSED(Id), void *UNUSED(Data))
{
    // Handle custom service requests from other plugins
    return false;
}

const char **cPluginXmltv2vdr::SVDRPHelpPages(void)
{
    // Returns help text
    static const char *HelpPages[]=
    {
        "UPDT\n"
        "    Start epg update",
    };
    return HelpPages;
}

cString cPluginXmltv2vdr::SVDRPCommand(const char *Command, const char *UNUSED(Option), int &ReplyCode)
{
    // Process SVDRP commands

    cString output;
    if (!strcasecmp(Command,"UPDT"))
    {
        if (!epgsources.Count())
        {
            ReplyCode=550;
            output="No epg sources installed\n";
        }
        else
        {
            if (executeepgsources())
            {
                ReplyCode=250;
                output="Update started\n";
            }
            else
            {
                ReplyCode=550;
                output="Update already running\n";
            }
        }
    }
    else
    {
        return NULL;
    }
    return output;
}

VDRPLUGINCREATOR(cPluginXmltv2vdr) // Don't touch this!
